import logging
from decimal import Decimal

import requests

from ..upow_transactions.helpers import string_to_point, round_up_decimal
from ..upow_transactions.transaction_input import TransactionInput


class WalletRepository:
    def __init__(self, node_url: str) -> None:
        self.node_url = node_url

    def get_address_info(
        self,
        address: str,
        stake_outputs: bool = False,
        delegate_spent_votes: bool = False,
        delegate_unspent_votes: bool = False,
        address_state: bool = False,
        inode_registration_outputs: bool = False,
        validator_unspent_votes: bool = False,
    ):
        request = requests.get(
            f"{self.node_url}/get_address_info",
            {
                "address": address,
                "transactions_count_limit": 0,
                "show_pending": True,
                "stake_outputs": stake_outputs,
                "delegate_spent_votes": delegate_spent_votes,
                "delegate_unspent_votes": delegate_unspent_votes,
                "address_state": address_state,
                "inode_registration_outputs": inode_registration_outputs,
                "validator_unspent_votes": validator_unspent_votes,
            },
        )
        request.raise_for_status()
        result = request.json()["result"]

        return result

    def get_dobby_info(self):
        request = requests.get(f"{self.node_url}/dobby_info")
        request.raise_for_status()
        result = request.json()["result"]
        return result

    def get_transaction(self, tx_hash: str):
        url = f"{self.node_url}/get_transaction?tx_hash={tx_hash}&verify=false"
        request = requests.get(url)
        request.raise_for_status()
        result = request.json()["result"]
        return result

    def get_validators_info(self, inode: str = None):
        params = {"inode": inode} if inode else {}
        request = requests.get(f"{self.node_url}/get_validators_info", params)
        request.raise_for_status()
        result = request.json()
        return result

    def get_inode_ballot_input_by_address_from_json(
        self,
        json,
        address,
        inode_address,
        pending_spent_outputs=None,
        check_pending_txs: bool = True,
    ):
        if pending_spent_outputs is None:
            pending_spent_outputs = []
        pending_spent_outputs = (
            [(output["tx_hash"], output["index"]) for output in pending_spent_outputs]
            if check_pending_txs is True
            else []
        )
        inode_ballot_inputs = []
        for validator_info in json:
            if validator_info["validator"] == address:
                for validator_voted_for in validator_info["vote"]:
                    if (
                        validator_voted_for["tx_hash"],
                        validator_voted_for["index"],
                    ) in pending_spent_outputs:
                        continue
                    if validator_voted_for["wallet"] != inode_address:
                        continue
                    tx_input = TransactionInput(
                        validator_voted_for["tx_hash"], validator_voted_for["index"]
                    )
                    tx_input.amount = Decimal(str(validator_voted_for["vote_count"]))
                    tx_input.public_key = string_to_point(address)
                    inode_ballot_inputs.append(tx_input)
        return inode_ballot_inputs

    def get_delegates_info(self, validator: str = None):
        params = {"validator": validator} if validator else {}
        request = requests.get(f"{self.node_url}/get_delegates_info", params)
        request.raise_for_status()
        result = request.json()
        return result

    def get_validator_ballot_input_by_address_from_json(
        self,
        json,
        address,
        validator_address,
        pending_spent_outputs=None,
        check_pending_txs: bool = True,
    ):
        if pending_spent_outputs is None:
            pending_spent_outputs = []
        pending_spent_outputs = (
            [(output["tx_hash"], output["index"]) for output in pending_spent_outputs]
            if check_pending_txs is True
            else []
        )
        validator_ballot_inputs = []
        for delegate_info in json:
            if delegate_info["delegate"] == address:
                for delegate_voted_for in delegate_info["vote"]:
                    if (
                        delegate_voted_for["tx_hash"],
                        delegate_voted_for["index"],
                    ) in pending_spent_outputs:
                        continue
                    if delegate_voted_for["wallet"] != validator_address:
                        continue
                    tx_input = TransactionInput(
                        delegate_voted_for["tx_hash"], delegate_voted_for["index"]
                    )
                    tx_input.amount = Decimal(str(delegate_voted_for["vote_count"]))
                    tx_input.public_key = string_to_point(address)
                    validator_ballot_inputs.append(tx_input)
        return validator_ballot_inputs

    def get_address_input_from_json(self, result, address):
        pending_spent_outputs = [
            (output["tx_hash"], output["index"])
            for output in result["pending_spent_outputs"]
        ]
        tx_inputs = []
        for spendable_tx_input in result["spendable_outputs"]:
            if (
                spendable_tx_input["tx_hash"],
                spendable_tx_input["index"],
            ) in pending_spent_outputs:
                continue
            tx_input = TransactionInput(
                spendable_tx_input["tx_hash"], spendable_tx_input["index"]
            )
            tx_input.amount = Decimal(str(spendable_tx_input["amount"]))
            tx_input.public_key = string_to_point(address)
            tx_inputs.append(tx_input)
        return tx_inputs

    def get_stake_input_from_json(
        self, result, address, check_pending_txs: bool = True
    ):
        pending_spent_outputs = (
            [
                (output["tx_hash"], output["index"])
                for output in result["pending_spent_outputs"]
            ]
            if check_pending_txs is True
            else []
        )
        stake_tx_input = []
        if result["stake_outputs"]:
            for stake_tx_output in result["stake_outputs"]:
                if (
                    stake_tx_output["tx_hash"],
                    stake_tx_output["index"],
                ) in pending_spent_outputs:
                    continue
                tx_input = TransactionInput(
                    stake_tx_output["tx_hash"], stake_tx_output["index"]
                )
                tx_input.amount = Decimal(str(stake_tx_output["amount"]))
                tx_input.public_key = string_to_point(address)
                stake_tx_input.append(tx_input)
        return stake_tx_input

    def get_inode_registration_input_from_json(self, json, address):
        pending_spent_outputs = [
            (output["tx_hash"], output["index"])
            for output in json["pending_spent_outputs"]
        ]
        inode_registration_input = []
        if json["inode_registration_outputs"]:
            for inode_reg_output in json["inode_registration_outputs"]:
                if (
                    inode_reg_output["tx_hash"],
                    inode_reg_output["index"],
                ) in pending_spent_outputs:
                    continue
                tx_input = TransactionInput(
                    inode_reg_output["tx_hash"], inode_reg_output["index"]
                )
                tx_input.amount = Decimal(str(inode_reg_output["amount"]))
                tx_input.public_key = string_to_point(address)
                inode_registration_input.append(tx_input)
        return inode_registration_input

    def get_delegate_spent_votes_from_json(self, json, check_pending_txs: bool = True):
        """
        Fetches the delegate_spent_votes data from the json and converts that data in TransactionInput.

        :param check_pending_txs: This indicates that to ignore the votes if the spent votes is in pending state
        :param json: Json data of address_info
        :return: The delegate_spent_votes in List[TransactionInput] of the account.
        """
        pending_spent_outputs = (
            [
                (output["tx_hash"], output["index"])
                for output in json["pending_spent_outputs"]
            ]
            if check_pending_txs is True
            else []
        )
        delegate_vote_tx_input = []
        if json["delegate_spent_votes"]:
            for delegate_spent_vote in json["delegate_spent_votes"]:
                if (
                    delegate_spent_vote["tx_hash"],
                    delegate_spent_vote["index"],
                ) in pending_spent_outputs:
                    continue
                tx_input = TransactionInput(
                    delegate_spent_vote["tx_hash"], delegate_spent_vote["index"]
                )
                tx_input.amount = Decimal(str(delegate_spent_vote["amount"]))
                delegate_vote_tx_input.append(tx_input)
        return delegate_vote_tx_input

    def get_delegate_unspent_votes_from_json(
        self, json, address: str = None, check_pending_txs: bool = True
    ):
        """
        Fetches the delegate_unspent_votes data from the json and converts that data in TransactionInput.

        :param address: Address of delegate
        :param check_pending_txs: This indicates that to ignore the votes if the spent votes is in pending state
        :param json: Json data of address_info
        :return: The delegate_unspent_votes in List[TransactionInput] of the account.
        """
        pending_spent_outputs = (
            [
                (output["tx_hash"], output["index"])
                for output in json["pending_spent_outputs"]
            ]
            if check_pending_txs is True
            else []
        )
        delegate_vote_tx_input = []
        if json["delegate_unspent_votes"]:
            for delegate_unspent_votes in json["delegate_unspent_votes"]:
                if (
                    delegate_unspent_votes["tx_hash"],
                    delegate_unspent_votes["index"],
                ) in pending_spent_outputs:
                    continue
                tx_input = TransactionInput(
                    delegate_unspent_votes["tx_hash"], delegate_unspent_votes["index"]
                )
                tx_input.amount = Decimal(str(delegate_unspent_votes["amount"]))
                tx_input.public_key = string_to_point(address) if address else None
                delegate_vote_tx_input.append(tx_input)
        return delegate_vote_tx_input

    def get_validator_unspent_votes_from_json(
        self, json, address, check_pending_txs: bool = True
    ):
        """
        Fetches the validator_unspent_votes data from the json and converts that data in TransactionInput.

        :param address: Address of validator
        :param check_pending_txs: This indicates that to ignore the votes if the spent votes is in pending state
        :param json: Json data of address_info
        :return: The validator_unspent_votes in List[TransactionInput] of the account.
        """
        pending_spent_outputs = (
            [
                (output["tx_hash"], output["index"])
                for output in json["pending_spent_outputs"]
            ]
            if check_pending_txs is True
            else []
        )
        validator_vote_tx_input = []
        if json["validator_unspent_votes"]:
            for validator_unspent_votes in json["validator_unspent_votes"]:
                if (
                    validator_unspent_votes["tx_hash"],
                    validator_unspent_votes["index"],
                ) in pending_spent_outputs:
                    continue
                tx_input = TransactionInput(
                    validator_unspent_votes["tx_hash"], validator_unspent_votes["index"]
                )
                tx_input.amount = Decimal(str(validator_unspent_votes["amount"]))
                tx_input.public_key = string_to_point(address)
                validator_vote_tx_input.append(tx_input)
        return validator_vote_tx_input

    def get_delegates_all_power(self, json):
        delegates_unspent_votes = self.get_delegate_unspent_votes_from_json(
            json, check_pending_txs=False
        )
        delegates_spent_votes = self.get_delegate_spent_votes_from_json(
            json, check_pending_txs=False
        )
        delegates_unspent_votes.extend(delegates_spent_votes)
        assert (
            sum(delegate_votes.amount for delegate_votes in delegates_unspent_votes)
            <= 10
        )
        return delegates_unspent_votes

    def get_pending_vote_as_delegate_transaction_from_json(self, address, json):
        pending_transactions = json["pending_transactions"]
        pending_vote_as_delegate_transaction = []
        for tx in pending_transactions:
            if (
                tx["transaction_type"] == "VOTE_AS_DELEGATE"
                and tx["inputs"][0]["address"] == address
            ):
                pending_vote_as_delegate_transaction.append(tx)
        return pending_vote_as_delegate_transaction

    def get_balance_info(self, address: str):
        """
        Fetches the account data from the node and calculates the pending balance.

        :param address: The address of the account.
        :return: The total balance and pending balance of the account.
        :raises: ConnectionError, ValueError, KeyError
        """
        try:
            # Send the request to the node
            request = requests.get(
                f"{self.node_url}/get_address_info",
                params={"address": address, "show_pending": True},
            )
            request.raise_for_status()  # Raises an HTTPError if the HTTP request returned an unsuccessful status code

            response = request.json()
            result = response.get("result")

            if not response.get("ok"):
                logging.error(response.get("error"))
                return None, None, None, None, True

            # Handle potential missing 'result' key
            if result is None:
                logging.error("Missing 'result' key in response")
                return None, None, None, None, True

            pending_transactions = result.get("pending_transactions", [])
            spendable_outputs = result.get("spendable_outputs", [])

            # Create a set of spendable transaction hashes for easy lookup
            spendable_hashes = {output["tx_hash"] for output in spendable_outputs}

            # Ensure the balance is a string before converting to Decimal
            total_balance = Decimal(str(result["balance"]))
            pending_balance = Decimal("0")
            stake_balance = Decimal(str(result["stake"]))
            pending_stake_balance = Decimal("0")

            for transaction in pending_transactions:
                # Adjust the balance based on inputs
                for input in transaction.get("inputs", []):
                    if (
                        input.get("address") == address
                        and input.get("tx_hash") in spendable_hashes
                    ):
                        input_amount = Decimal(str(input.get("amount", "0")))
                        if any(
                            tx_output.get("type") == "UN_STAKE"
                            for tx_output in transaction.get("outputs", [])
                        ):
                            pending_balance += input_amount
                        elif transaction.get("transaction_type") == "REGULAR":
                            pending_balance -= input_amount

                # Adjust the balance based on outputs
                for output in transaction.get("outputs", []):
                    if output.get("address") == address:
                        output_amount = Decimal(str(output.get("amount", "0")))
                        if output.get("type") == "STAKE":
                            pending_stake_balance += output_amount
                        elif output.get("type") == "UN_STAKE":
                            pending_stake_balance -= output_amount
                        elif output.get("type") == "REGULAR":
                            pending_balance += output_amount

            # Format the total balance and pending balance to remove unnecessary trailing zeros
            formatted_total_balance = round_up_decimal(total_balance)
            formatted_pending_balance = round_up_decimal(pending_balance)
            formatted_pending_stake_balance = round_up_decimal(pending_stake_balance)
            formatted_stake_balance = round_up_decimal(stake_balance)

            balance_data = (
                formatted_total_balance,
                formatted_pending_balance,
                formatted_stake_balance,
                formatted_pending_stake_balance,
                False,
            )
            return balance_data

        except requests.RequestException as e:
            # Handles exceptions that occur during the request
            logging.error(f"Error during request to node: {e}")
            return None, None, None, None, True

        except ValueError as e:
            # Handles JSON decoding errors
            logging.error(f"Error decoding JSON response: {e}")
            return None, None, None, None, True

        except KeyError as e:
            # Handles missing keys in response data
            logging.error(f"Missing expected data in response: {e}")
            return None, None, None, None, True



from decimal import Decimal
from .constants import ENDIAN
from .helpers import InputType, sha256
from .transaction_output import TransactionOutput


class CoinbaseTransaction:
    _hex: str = None

    def __init__(self, block_hash: str, address: str, amount: Decimal):
        self.block_hash = block_hash
        self.address = address
        self.amount = amount
        self.outputs = [TransactionOutput(address, amount)]

    def hex(self):
        if self._hex is not None:
            return self._hex
        hex_inputs = (
            bytes.fromhex(self.block_hash) + (0).to_bytes(1, ENDIAN)
        ).hex() + InputType.REGULAR.value.to_bytes(1, ENDIAN).hex()
        hex_outputs = "".join(tx_output.tobytes().hex() for tx_output in self.outputs)

        if all(len(tx_output.address_bytes) == 64 for tx_output in self.outputs):
            version = 1
        elif all(len(tx_output.address_bytes) == 33 for tx_output in self.outputs):
            version = 2
        else:
            raise NotImplementedError()

        self._hex = "".join(
            [
                version.to_bytes(1, ENDIAN).hex(),
                (1).to_bytes(1, ENDIAN).hex(),
                hex_inputs,
                (len(self.outputs)).to_bytes(1, ENDIAN).hex(),
                hex_outputs,
                (36).to_bytes(1, ENDIAN).hex(),
            ]
        )

        return self._hex

    def hash(self):
        return sha256(self.hex())



from decimal import Decimal
from typing import Tuple
from fastecdsa import ecdsa
from fastecdsa.point import Point
from .constants import CURVE, ENDIAN, SMALLEST
from .helpers import point_to_string, string_to_point, InputType


class TransactionInput:
    public_key = None

    signed: Tuple[int, int] = None
    amount: Decimal = None

    def __init__(
        self,
        input_tx_hash: str,
        index: int,
        private_key: int = None,
        transaction=None,
        amount: Decimal = None,
        public_key: Point = None,
        input_type: InputType = InputType.REGULAR,
    ):
        self.tx_hash = input_tx_hash
        self.index = index
        self.private_key = private_key
        self.transaction = transaction
        self.transaction_info = None
        self.amount = amount
        self.public_key = public_key
        self.input_type = input_type
        if transaction is not None and amount is None:
            self.get_related_output()

    async def get_transaction(self):
        return self.transaction

    async def get_transaction_info(self):
        assert self.transaction_info is not None
        return self.transaction_info

    async def get_related_output(self):
        tx = await self.get_transaction()
        related_output = tx.outputs[self.index]
        self.amount = related_output.amount
        return related_output

    async def get_related_input(self):
        tx = await self.get_transaction()
        related_input = tx.inputs[0]
        return related_input

    async def get_related_input_info(self):
        tx = await self.get_transaction_info()
        related_input = {"address": tx["inputs_addresses"][0]}
        return related_input

    async def get_related_output_info(self):
        tx = await self.get_transaction_info()
        related_output = {
            "address": tx["outputs_addresses"][self.index],
            "amount": Decimal(tx["outputs_amounts"][self.index]) / SMALLEST,
        }
        self.amount = related_output["amount"]
        return related_output

    async def get_amount(self):
        if self.amount is None:
            if self.transaction is not None:
                return self.transaction.outputs[self.index].amount
            else:
                await self.get_related_output_info()
        return self.amount

    async def get_address(self):
        if self.transaction is not None:
            return (await self.get_related_output()).address
        return (await self.get_related_output_info())["address"]

    async def get_voter_address(self):
        if self.transaction is not None:
            return (await self.get_related_input()).address
        return (await self.get_related_input_info())["address"]

    def sign(self, tx_hex: str, private_key: int = None):
        private_key = private_key if private_key is not None else self.private_key
        self.signed = ecdsa.sign(bytes.fromhex(tx_hex), private_key)

    async def get_public_key(self):
        return self.public_key or string_to_point(await self.get_address())

    async def get_voter_public_key(self):
        return self.public_key or string_to_point(await self.get_voter_address())

    def tobytes(self):
        return (
            bytes.fromhex(self.tx_hash)
            + self.index.to_bytes(1, ENDIAN)
            + self.input_type.to_bytes(1, ENDIAN)
        )

    def get_signature(self):
        return (
            self.signed[0].to_bytes(32, ENDIAN).hex()
            + self.signed[1].to_bytes(32, ENDIAN).hex()
        )

    async def verify(self, input_tx) -> bool:
        try:
            public_key = await self.get_public_key()
        except AssertionError:
            return False
        # print('verifying with', point_to_string(public_key))

        return ecdsa.verify(
            self.signed, bytes.fromhex(input_tx), public_key, CURVE
        ) or ecdsa.verify(self.signed, input_tx, public_key, CURVE)

    async def verify_revoke_tx(self, input_tx) -> bool:
        try:
            public_key = await self.get_voter_public_key()
        except AssertionError:
            return False
        # print('verifying with', point_to_string(public_key))

        return ecdsa.verify(
            self.signed, bytes.fromhex(input_tx), public_key, CURVE
        ) or ecdsa.verify(self.signed, input_tx, public_key, CURVE)

    @property
    def as_dict(self):
        self_dict = vars(self).copy()
        self_dict["signed"] = self_dict["signed"] is not None
        if "public_key" in self_dict:
            self_dict["public_key"] = point_to_string(self_dict["public_key"])
        if "transaction" in self_dict:
            del self_dict["transaction"]
        if "private_key" in self_dict:
            del self_dict["private_key"]
        return self_dict

    def __eq__(self, other):
        assert isinstance(other, self.__class__)
        return (self.tx_hash, self.index) == (other.tx_hash, other.index)



from decimal import Decimal
from .constants import ENDIAN, SMALLEST, CURVE
from .helpers import (
    byte_length,
    string_to_point,
    string_to_bytes,
    OutputType,
)


class TransactionOutput:
    def __init__(
        self,
        address: str,
        amount: Decimal,
        transaction_type: OutputType = OutputType.REGULAR,
    ):
        from fastecdsa.point import Point

        if isinstance(address, Point):
            raise Exception(
                "TransactionOutput does not accept Point anymore. Pass the address string instead"
            )
        self.address = address
        self.address_bytes = string_to_bytes(address)
        self.public_key = string_to_point(address)
        assert (amount * SMALLEST) % 1 == 0.0, "too many decimal digits"
        self.amount = amount
        self.transaction_type = transaction_type
        self.is_stake = transaction_type == OutputType.STAKE

    def tobytes(self):
        amount = int(self.amount * SMALLEST)
        count = byte_length(amount)
        return (
            self.address_bytes
            + count.to_bytes(1, ENDIAN)
            + amount.to_bytes(count, ENDIAN)
            + self.transaction_type.to_bytes(1, ENDIAN)
        )

    def verify(self):
        return self.amount > 0 and CURVE.is_point_on_curve(
            (self.public_key.x, self.public_key.y)
        )

    @property
    def as_dict(self):
        res = vars(self).copy()
        if "public_key" in res:
            del res["public_key"]
        return res



from decimal import Decimal
from io import BytesIO
from typing import List
from fastecdsa import keys
from icecream import ic
from .coinbase_transaction import CoinbaseTransaction
from .constants import ENDIAN, SMALLEST, CURVE
from .helpers import (
    get_transaction_type_from_message,
    sha256,
    TransactionType,
    OutputType,
    bytes_to_string,
    point_to_string,
    InputType,
)
from .transaction_input import TransactionInput
from .transaction_output import TransactionOutput

print = ic


class Transaction:
    def __init__(
        self,
        inputs: List[TransactionInput],
        outputs: List[TransactionOutput],
        message: bytes = None,
        version: int = None,
    ):
        if len(inputs) >= 256:
            raise Exception(
                f"You can spend max 255 inputs in a single transactions, not {len(inputs)}"
            )
        if len(outputs) >= 256:
            raise Exception(
                f"You can have max 255 outputs in a single transactions, not {len(outputs)}"
            )
        self.inputs = inputs
        self.outputs = outputs
        self.message = message
        self.transaction_type = get_transaction_type_from_message(message)
        if version is None:
            if all(len(tx_output.address_bytes) == 64 for tx_output in outputs):
                version = 1
            elif all(len(tx_output.address_bytes) == 33 for tx_output in outputs):
                version = 3
            else:
                raise NotImplementedError()
        if version > 3:
            raise NotImplementedError()
        self.version = version
        self._hex: str = None
        self.fees: Decimal = None
        self.tx_hash: str = None

    def hex(self, full: bool = True):
        inputs, outputs = self.inputs, self.outputs
        hex_inputs = "".join(tx_input.tobytes().hex() for tx_input in inputs)
        hex_outputs = "".join(tx_output.tobytes().hex() for tx_output in outputs)

        version = self.version

        self._hex = "".join(
            [
                version.to_bytes(1, ENDIAN).hex(),
                len(inputs).to_bytes(1, ENDIAN).hex(),
                hex_inputs,
                (len(outputs)).to_bytes(1, ENDIAN).hex(),
                hex_outputs,
            ]
        )

        if not full and (version <= 2 or self.message is None):
            return self._hex

        if self.message is not None:
            if version <= 2:
                self._hex += bytes([1, len(self.message)]).hex()
            else:
                self._hex += bytes([1]).hex()
                self._hex += (len(self.message)).to_bytes(2, ENDIAN).hex()
            self._hex += self.message.hex()
            if not full:
                return self._hex
        else:
            self._hex += (0).to_bytes(1, ENDIAN).hex()

        signatures = []
        for tx_input in inputs:
            signed = tx_input.get_signature()
            if signed not in signatures:
                signatures.append(signed)
                self._hex += signed

        return self._hex

    def hash(self):
        if self.tx_hash is None:
            self.tx_hash = sha256(self.hex())
        return self.tx_hash

    def _verify_double_spend_same_transaction(self):
        used_inputs = []
        for tx_input in self.inputs:
            input_hash = f"{tx_input.tx_hash}{tx_input.index}"
            if input_hash in used_inputs:
                return False
            used_inputs.append(input_hash)
        return True

    async def _check_signature(self):
        tx_hex = self.hex(False)
        checked_signatures = []
        for tx_input in self.inputs:
            if tx_input.signed is None:
                print("not signed")
                return False
            await tx_input.get_public_key()
            signature = (tx_input.public_key, tx_input.signed)
            if signature in checked_signatures:
                continue
            if not await tx_input.verify(tx_hex):
                print("signature not valid")
                return False
            checked_signatures.append(signature)
        return True

    def sign(self, private_keys: list = []):
        for private_key in private_keys:
            for input in self.inputs:
                if input.private_key is None and (
                    input.public_key or input.transaction
                ):
                    public_key = keys.get_public_key(private_key, CURVE)
                    input_public_key = (
                        input.public_key
                        or input.transaction.outputs[input.index].public_key
                    )
                    if public_key == input_public_key:
                        input.private_key = private_key
        for input in self.inputs:
            if input.private_key is not None:
                input.sign(self.hex(False))
        return self

    @staticmethod
    async def from_hex(hexstring: str, check_signatures: bool = True):
        tx_bytes = BytesIO(bytes.fromhex(hexstring))
        version = int.from_bytes(tx_bytes.read(1), ENDIAN)
        if version > 3:
            raise NotImplementedError()

        inputs_count = int.from_bytes(tx_bytes.read(1), ENDIAN)

        inputs = []

        for i in range(0, inputs_count):
            tx_hex = tx_bytes.read(32).hex()
            tx_index = int.from_bytes(tx_bytes.read(1), ENDIAN)
            input_type = int.from_bytes(tx_bytes.read(1), ENDIAN)
            inputs.append(
                TransactionInput(
                    tx_hex, index=tx_index, input_type=InputType(input_type)
                )
            )

        outputs_count = int.from_bytes(tx_bytes.read(1), ENDIAN)

        outputs = []

        for i in range(0, outputs_count):
            pubkey = tx_bytes.read(64 if version == 1 else 33)
            amount_length = int.from_bytes(tx_bytes.read(1), ENDIAN)
            amount = int.from_bytes(tx_bytes.read(amount_length), ENDIAN) / Decimal(
                SMALLEST
            )
            transaction_type = int.from_bytes(tx_bytes.read(1), ENDIAN)
            outputs.append(
                TransactionOutput(
                    bytes_to_string(pubkey), amount, OutputType(transaction_type)
                )
            )

        specifier = int.from_bytes(tx_bytes.read(1), ENDIAN)
        if specifier == 36:
            # assert len(inputs) == 1 and len(outputs) == 1
            assert len(inputs) == 1
            coinbase_transaction = CoinbaseTransaction(
                inputs[0].tx_hash, outputs[0].address, outputs[0].amount
            )
            if len(outputs) > 1:
                coinbase_transaction.outputs.extend(outputs[1:])
            return coinbase_transaction
        else:
            if specifier == 1:
                message_length = int.from_bytes(
                    tx_bytes.read(1 if version <= 2 else 2), ENDIAN
                )
                message = tx_bytes.read(message_length)
            else:
                message = None
                assert specifier == 0

            signatures = []

            while True:
                signed = (
                    int.from_bytes(tx_bytes.read(32), ENDIAN),
                    int.from_bytes(tx_bytes.read(32), ENDIAN),
                )
                if signed[0] == 0:
                    break
                signatures.append(signed)

            if len(signatures) == 1:
                for tx_input in inputs:
                    tx_input.signed = signatures[0]
            elif len(inputs) == len(signatures):
                for i, tx_input in enumerate(inputs):
                    tx_input.signed = signatures[i]
            else:
                if not check_signatures:
                    return Transaction(inputs, outputs, message, version)
                index = {}
                for tx_input in inputs:
                    public_key = point_to_string(await tx_input.get_public_key())
                    if public_key not in index.keys():
                        index[public_key] = []
                    index[public_key].append(tx_input)

                for i, signed in enumerate(signatures):
                    for tx_input in index[list(index.keys())[i]]:
                        tx_input.signed = signed

            return Transaction(inputs, outputs, message, version)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.hex() == other.hex()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)



from decimal import Decimal

from fastecdsa import keys

from ..utils.repository import WalletRepository
from ..upow_transactions.constants import CURVE, MAX_INODES
from ..upow_transactions.helpers import (
    point_to_string,
    OutputType,
    TransactionType,
)
from ..upow_transactions.transaction import Transaction
from ..upow_transactions.transaction_output import TransactionOutput


class Utils:
    def __init__(self, node_url=None):
        self.NODE_URL = node_url
        self.repo = WalletRepository(self.NODE_URL)

    def set_node_url(self, new_url: str):
        if new_url != self.NODE_URL:
            self.NODE_URL = new_url
            self.repo = WalletRepository(self.NODE_URL)

    def get_balance_info(self, address: str):
        result = self.repo.get_balance_info(address)
        return result

    def get_tx(self, hash: str):
        result = self.repo.get_transaction(hash)
        return result

    async def create_transaction(
        self,
        private_key,
        receiving_address,
        amount,
        message: bytes = None,
        send_back_address=None,
    ):
        amount = Decimal(amount)
        inputs = []
        sender_address = point_to_string(keys.get_public_key(private_key, CURVE))
        if send_back_address is None:
            send_back_address = sender_address

        r_json = self.repo.get_address_info(sender_address)
        address_inputs = self.repo.get_address_input_from_json(
            r_json, address=sender_address
        )
        inputs.extend(address_inputs)
        if not inputs:
            raise Exception("No spendable outputs")

        if sum(input.amount for input in inputs) < amount:
            raise Exception(f"Error: You don't have enough funds")

        transaction_inputs = self.select_transaction_input(inputs, amount)

        transaction_amount = sum(input.amount for input in transaction_inputs)

        transaction = Transaction(
            transaction_inputs,
            [TransactionOutput(receiving_address, amount=amount)],
            message,
        )
        if transaction_amount > amount:
            transaction.outputs.append(
                TransactionOutput(send_back_address, transaction_amount - amount)
            )

        transaction.sign([private_key])

        return transaction

    async def create_transaction_to_send_multiple_wallet(
        self,
        private_key,
        receiving_addresses,
        amounts,
        message: bytes = None,
        send_back_address=None,
    ):
        if len(receiving_addresses) != len(amounts):
            raise Exception(
                "Receiving addresses length is different from amounts length"
            )
        amounts = [Decimal(amount) for amount in amounts]
        total_amount = sum(amounts)
        total_amount = Decimal(total_amount)
        inputs = []
        sender_address = point_to_string(keys.get_public_key(private_key, CURVE))
        if send_back_address is None:
            send_back_address = sender_address

        r_json = self.repo.get_address_info(sender_address)
        address_inputs = self.repo.get_address_input_from_json(
            r_json, address=sender_address
        )
        inputs.extend(address_inputs)

        if not inputs:
            raise Exception("No spendable outputs")

        total_input_amount = sum(input.amount for input in inputs)

        if total_input_amount < total_amount:
            raise Exception(f"Error: You don't have enough funds")

        transaction_inputs = []
        transaction_outputs = []

        # Select inputs to cover the total amount
        input_amount = Decimal(0)
        for tx_input in sorted(inputs, key=lambda item: item.amount, reverse=True):
            transaction_inputs.append(tx_input)
            input_amount += tx_input.amount
            if input_amount >= total_amount:
                break

        # Create outputs for each receiving address
        for receiving_address, amount in zip(receiving_addresses, amounts):
            transaction_outputs.append(
                TransactionOutput(receiving_address, amount=Decimal(amount))
            )

        # If there's change, add an output back to the sender
        change_amount = input_amount - total_amount
        if change_amount > 0:
            transaction_outputs.append(
                TransactionOutput(send_back_address, amount=change_amount)
            )

        transaction = Transaction(transaction_inputs, transaction_outputs, message)
        transaction.sign([private_key])

        return transaction

    async def create_stake_transaction(
        self, private_key, amount, send_back_address=None
    ):
        amount = Decimal(amount)
        inputs = []

        sender_address = point_to_string(keys.get_public_key(private_key, CURVE))
        if send_back_address is None:
            send_back_address = sender_address

        result_json = self.repo.get_address_info(
            sender_address,
            stake_outputs=True,
            delegate_unspent_votes=True,
            delegate_spent_votes=True,
        )
        inputs.extend(
            self.repo.get_address_input_from_json(result_json, address=sender_address)
        )

        if not inputs:
            raise Exception("No spendable outputs")

        if sum(input.amount for input in inputs) < amount:
            raise Exception(f"Error: You don't have enough funds")

        stake_inputs = self.repo.get_stake_input_from_json(
            result_json, address=sender_address, check_pending_txs=False
        )
        if stake_inputs:
            raise Exception("Already staked")

        transaction_inputs = []

        for tx_input in sorted(inputs, key=lambda item: item.amount):
            if tx_input.amount >= amount:
                transaction_inputs.append(tx_input)
                break
        for tx_input in sorted(inputs, key=lambda item: item.amount, reverse=True):
            if sum(input.amount for input in transaction_inputs) >= amount:
                break
            transaction_inputs.append(tx_input)

        transaction_amount = sum(input.amount for input in transaction_inputs)

        transaction = Transaction(
            transaction_inputs,
            [
                TransactionOutput(
                    sender_address, amount=amount, transaction_type=OutputType.STAKE
                )
            ],
        )

        if transaction_amount > amount:
            transaction.outputs.append(
                TransactionOutput(send_back_address, transaction_amount - amount)
            )

        if not self.repo.get_delegates_all_power(result_json):
            voting_power = Decimal(10)
            transaction.outputs.append(
                TransactionOutput(
                    sender_address,
                    voting_power,
                    transaction_type=OutputType.DELEGATE_VOTING_POWER,
                )
            )

        transaction.sign([private_key])

        return transaction

    async def create_unstake_transaction(self, private_key):
        sender_address = point_to_string(keys.get_public_key(private_key, CURVE))
        result_json = self.repo.get_address_info(
            sender_address, stake_outputs=True, delegate_spent_votes=True
        )
        stake_inputs = self.repo.get_stake_input_from_json(
            result_json, address=sender_address
        )
        if not stake_inputs:
            raise Exception(f"Error: There is nothing staked")
        amount = stake_inputs[0].amount

        if self.repo.get_delegate_spent_votes_from_json(
            result_json, check_pending_txs=False
        ):
            raise Exception("Kindly release the votes.")

        pending_vote_tx = self.repo.get_pending_vote_as_delegate_transaction_from_json(
            sender_address, result_json
        )
        if pending_vote_tx:
            raise Exception("Kindly release the votes. Vote transaction is in pending")

        transaction = Transaction(
            [stake_inputs[0]],
            [
                TransactionOutput(
                    sender_address, amount=amount, transaction_type=OutputType.UN_STAKE
                )
            ],
        )
        transaction.sign([private_key])
        return transaction

    async def create_inode_registration_transaction(self, private_key):
        amount = Decimal(1000)
        inputs = []
        address = point_to_string(keys.get_public_key(private_key, CURVE))

        result_json = self.repo.get_address_info(
            address, stake_outputs=True, address_state=True
        )
        inputs.extend(
            self.repo.get_address_input_from_json(result_json, address=address)
        )

        if not inputs:
            raise Exception("No spendable outputs")

        if sum(input.amount for input in inputs) < amount:
            raise Exception(f"Error: You don't have enough funds")

        stake_inputs = self.repo.get_stake_input_from_json(result_json, address=address)
        if not stake_inputs:
            raise Exception(f"You are not a delegate. Become a delegate by staking.")

        is_inode_registered = result_json["is_inode"]
        if is_inode_registered:
            raise Exception(f"This address is already registered as inode.")

        is_validator_registered = result_json["is_validator"]
        if is_validator_registered:
            raise Exception(
                f"This address is registered as validator and a validator cannot be an inode."
            )

        inode_addresses = self.repo.get_dobby_info()
        if len(inode_addresses) >= MAX_INODES:
            raise Exception(f"{MAX_INODES} inodes are already registered.")

        transaction_inputs = []

        for tx_input in sorted(inputs, key=lambda item: item.amount):
            if tx_input.amount >= amount:
                transaction_inputs.append(tx_input)
                break
        for tx_input in sorted(inputs, key=lambda item: item.amount, reverse=True):
            if sum(input.amount for input in transaction_inputs) >= amount:
                break
            transaction_inputs.append(tx_input)

        transaction_amount = sum(input.amount for input in transaction_inputs)

        transaction = Transaction(
            transaction_inputs,
            [
                TransactionOutput(
                    address,
                    amount=amount,
                    transaction_type=OutputType.INODE_REGISTRATION,
                )
            ],
        )
        if transaction_amount > amount:
            transaction.outputs.append(
                TransactionOutput(address, transaction_amount - amount)
            )

        transaction.sign([private_key])
        return transaction

    async def create_inode_de_registration_transaction(self, private_key):
        inputs = []
        address = point_to_string(keys.get_public_key(private_key, CURVE))

        result_json = self.repo.get_address_info(
            address, inode_registration_outputs=True
        )
        inputs.extend(
            self.repo.get_inode_registration_input_from_json(
                result_json, address=address
            )
        )

        if not inputs:
            raise Exception("This address is not registered as an inode.")

        active_inode_addresses = self.repo.get_dobby_info()
        is_inode_active = any(
            entry.get("wallet") == address for entry in active_inode_addresses
        )
        if is_inode_active:
            raise Exception("This address is an active inode. Cannot de-register.")

        amount = inputs[0].amount
        message = self.string_to_bytes(str(TransactionType.INODE_DE_REGISTRATION.value))
        transaction = Transaction(
            inputs, [TransactionOutput(address, amount=amount)], message
        )
        transaction.sign([private_key])
        return transaction

    async def create_validator_registration_transaction(self, private_key):
        amount = Decimal(100)
        inputs = []
        address = point_to_string(keys.get_public_key(private_key, CURVE))
        result_json = self.repo.get_address_info(
            address, stake_outputs=True, address_state=True
        )
        inputs.extend(
            self.repo.get_address_input_from_json(result_json, address=address)
        )

        if not inputs:
            raise Exception("No spendable outputs")

        if sum(input.amount for input in inputs) < amount:
            raise Exception(f"Error: You don't have enough funds")

        stake_inputs = self.repo.get_stake_input_from_json(result_json, address=address)
        if not stake_inputs:
            raise Exception(f"You are not a delegate. Become a delegate by staking.")

        is_validator_registered = result_json["is_validator"]
        if is_validator_registered:
            raise Exception(f"This address is already registered as validator.")

        is_inode_registered = result_json["is_inode"]
        if is_inode_registered:
            raise Exception(
                f"This address is registered as inode and an inode cannot be a validator."
            )

        transaction_inputs = self.select_transaction_input(inputs, amount)

        transaction_amount = sum(input.amount for input in transaction_inputs)

        message = self.string_to_bytes(
            str(TransactionType.VALIDATOR_REGISTRATION.value)
        )
        transaction = Transaction(
            transaction_inputs,
            [
                TransactionOutput(
                    address,
                    amount=amount,
                    transaction_type=OutputType.VALIDATOR_REGISTRATION,
                )
            ],
            message,
        )

        voting_power = Decimal(10)
        transaction.outputs.append(
            TransactionOutput(
                address,
                voting_power,
                transaction_type=OutputType.VALIDATOR_VOTING_POWER,
            )
        )

        if transaction_amount > amount:
            transaction.outputs.append(
                TransactionOutput(address, transaction_amount - amount)
            )

        transaction.sign([private_key])
        return transaction

    async def create_voting_transaction(
        self, private_key, vote_range, vote_receiving_address
    ):
        try:
            vote_range = int(vote_range)
        except:
            raise Exception("Invalid voting range")
        if vote_range > 10:
            raise Exception("Voting should be in range of 10")
        if vote_range <= 0:
            raise Exception("Invalid voting range")

        address = point_to_string(keys.get_public_key(private_key, CURVE))
        result_json = self.repo.get_address_info(
            address,
            stake_outputs=True,
            address_state=True,
            validator_unspent_votes=True,
            delegate_unspent_votes=True,
        )
        stake_inputs = self.repo.get_stake_input_from_json(result_json, address=address)

        is_inode_registered = result_json["is_inode"]
        if is_inode_registered:
            raise Exception(f"This address is registered as inode. Cannot vote.")

        is_validator_registered = result_json["is_validator"]
        if is_validator_registered:
            return await self.vote_as_validator(
                private_key, vote_range, vote_receiving_address, result_json
            )
        elif stake_inputs:
            return await self.vote_as_delegate(
                private_key, vote_range, vote_receiving_address, result_json
            )
        else:
            raise Exception("Not eligible to vote")

    async def vote_as_validator(
        self, private_key, vote_range, vote_receiving_address, result_json
    ):
        address = point_to_string(keys.get_public_key(private_key, CURVE))
        vote_range = Decimal(vote_range)
        inputs = []
        inputs.extend(
            self.repo.get_validator_unspent_votes_from_json(result_json, address)
        )
        if not inputs:
            raise Exception("No voting outputs")

        if sum(input.amount for input in inputs) < vote_range:
            raise Exception(
                f"Error: You don't have enough voting power left. Kindly revoke some voting power."
            )

        transaction_inputs = self.select_transaction_input(inputs, vote_range)

        transaction_vote_range = sum(input.amount for input in transaction_inputs)

        message = self.string_to_bytes(str(TransactionType.VOTE_AS_VALIDATOR.value))
        transaction = Transaction(
            transaction_inputs,
            [
                TransactionOutput(
                    vote_receiving_address,
                    amount=vote_range,
                    transaction_type=OutputType.VOTE_AS_VALIDATOR,
                )
            ],
            message,
        )
        if transaction_vote_range > vote_range:
            transaction.outputs.append(
                TransactionOutput(
                    address,
                    transaction_vote_range - vote_range,
                    transaction_type=OutputType.VALIDATOR_VOTING_POWER,
                )
            )

        transaction.sign([private_key])
        return transaction

    async def vote_as_delegate(
        self, private_key, vote_range, vote_receiving_address, result_json
    ):
        address = point_to_string(keys.get_public_key(private_key, CURVE))

        vote_range = Decimal(vote_range)
        inputs = []
        inputs.extend(
            self.repo.get_delegate_unspent_votes_from_json(result_json, address)
        )
        if not inputs:
            raise Exception("No voting outputs")

        if sum(input.amount for input in inputs) < vote_range:
            raise Exception(
                f"Error: You don't have enough voting power left. Kindly release some voting power."
            )

        transaction_inputs = self.select_transaction_input(inputs, vote_range)

        transaction_vote_range = sum(input.amount for input in transaction_inputs)

        message = self.string_to_bytes(str(TransactionType.VOTE_AS_DELEGATE.value))
        transaction = Transaction(
            transaction_inputs,
            [
                TransactionOutput(
                    vote_receiving_address,
                    amount=vote_range,
                    transaction_type=OutputType.VOTE_AS_DELEGATE,
                )
            ],
            message,
        )
        if transaction_vote_range > vote_range:
            transaction.outputs.append(
                TransactionOutput(
                    address,
                    transaction_vote_range - vote_range,
                    transaction_type=OutputType.DELEGATE_VOTING_POWER,
                )
            )

        transaction.sign([private_key])
        return transaction

    async def create_revoke_transaction(self, private_key, revoke_from_address):
        address = point_to_string(keys.get_public_key(private_key, CURVE))
        result_json = self.repo.get_address_info(
            address, stake_outputs=True, address_state=True
        )

        stake_inputs = self.repo.get_stake_input_from_json(result_json, address=address)

        is_validator_registered = result_json["is_validator"]
        if is_validator_registered:
            return await self.revoke_vote_as_validator(
                private_key, revoke_from_address, result_json
            )
        elif stake_inputs:
            pass
            return await self.revoke_vote_as_delegate(
                private_key, revoke_from_address, result_json
            )
        else:
            raise Exception("Not eligible to revoke")

    async def revoke_vote_as_validator(self, private_key, inode_address, address_info):
        address = point_to_string(keys.get_public_key(private_key, CURVE))
        inode_ballot = self.repo.get_validators_info(inode_address)
        inode_ballot_inputs = self.repo.get_inode_ballot_input_by_address_from_json(
            inode_ballot,
            address,
            inode_address,
            pending_spent_outputs=address_info["pending_spent_outputs"],
        )
        if not inode_ballot_inputs:
            raise Exception("You have not voted.")

        message = self.string_to_bytes(str(TransactionType.REVOKE_AS_VALIDATOR.value))
        sum_of_votes = sum(
            inode_ballot_input.amount for inode_ballot_input in inode_ballot_inputs
        )
        transaction = Transaction(
            inode_ballot_inputs,
            [
                TransactionOutput(
                    address,
                    amount=sum_of_votes,
                    transaction_type=OutputType.VALIDATOR_VOTING_POWER,
                )
            ],
            message,
        )
        transaction.sign([private_key])
        return transaction

    async def revoke_vote_as_delegate(
        self, private_key, validator_address, address_info
    ):
        address = point_to_string(keys.get_public_key(private_key, CURVE))

        validator_ballot = self.repo.get_delegates_info(validator_address)
        validator_ballot_inputs = (
            self.repo.get_validator_ballot_input_by_address_from_json(
                validator_ballot,
                address,
                validator_address,
                pending_spent_outputs=address_info["pending_spent_outputs"],
            )
        )

        if not validator_ballot_inputs:
            raise Exception("You have not voted.")

        message = self.string_to_bytes(str(TransactionType.REVOKE_AS_DELEGATE.value))
        sum_of_votes = sum(
            validator_ballot_input.amount
            for validator_ballot_input in validator_ballot_inputs
        )
        transaction = Transaction(
            validator_ballot_inputs,
            [
                TransactionOutput(
                    address,
                    amount=sum_of_votes,
                    transaction_type=OutputType.DELEGATE_VOTING_POWER,
                )
            ],
            message,
        )
        transaction.sign([private_key])
        return transaction

    def select_transaction_input(self, inputs, amount):
        transaction_inputs = []
        for tx_input in sorted(inputs, key=lambda item: item.amount):
            if tx_input.amount >= amount:
                transaction_inputs.append(tx_input)
                break
        for tx_input in sorted(inputs, key=lambda item: item.amount, reverse=True):
            if sum(input.amount for input in transaction_inputs) >= amount:
                break
            transaction_inputs.append(tx_input)
        return transaction_inputs

    def string_to_bytes(self, string: str) -> bytes:
        if string is None:
            return None
        try:
            return bytes.fromhex(string)
        except ValueError:
            return string.encode("utf-8")



# upowpy/__init__.py
from .push import (
    send_transaction,
    stake_transaction,
    unstake_transaction,
    register_inode_transaction,
    deregister_inode_transaction,
    register_validator_transaction,
    vote_transaction,
    revoke_transaction,
    get_balance,
    find_tx,
    publickey_to_address,
    address_to_publickey,
    create_wallet,
)

__all__ = [
    "send_transaction",
    "stake_transaction",
    "unstake_transaction",
    "register_inode_transaction",
    "deregister_inode_transaction",
    "register_validator_transaction",
    "vote_transaction",
    "revoke_transaction",
    "get_balance",
    "find_tx",
    "publickey_to_address",
    "address_to_publickey",
    "create_wallet",
]



import logging
import requests
from .upow_transactions.helpers import (
    sha256,
    string_to_bytes,
    point_to_string,
    string_to_point,
)
from .utils.utils import Utils
import decimal
from fastecdsa import keys
from .upow_transactions.constants import CURVE

wallet_utils: Utils = Utils()


async def push_tx(tx, wallet_utils: Utils):
    try:
        r = requests.get(
            f"{wallet_utils.NODE_URL}/push_tx", {"tx_hex": tx.hex()}, timeout=10
        )
        r.raise_for_status()
        res = r.json()
        if res["ok"]:
            transaction_hash = sha256(tx.hex())
            return None, transaction_hash
        else:
            return "Transaction not pushed", None
    except Exception as e:
        logging.error(f"Error during request to node: {e}")
        return str(e), None


async def send_transaction(
    wallet_utils: Utils, private_key_hex, recipients, amounts, message=None
):
    private_key = int(private_key_hex, 16)
    recipients_list = recipients.split(",")
    amounts_list = amounts.split(",")
    message_bytes = string_to_bytes(message) if message else None
    if (
        len(recipients_list) > 1
        and len(amounts_list) > 1
        and len(recipients_list) == len(amounts_list)
    ):
        tx = await wallet_utils.create_transaction_to_send_multiple_wallet(
            private_key, recipients_list, amounts_list, message_bytes
        )
    else:
        receiver = recipients_list[0]
        amount = amounts_list[0]
        tx = await wallet_utils.create_transaction(
            private_key, receiver, amount, message_bytes
        )
    error_message, transaction_hash = await push_tx(tx, wallet_utils)
    if transaction_hash:
        return transaction_hash
    else:
        raise Exception(error_message)


async def stake_transaction(wallet_utils: Utils, private_key_hex, amount):
    try:
        private_key = int(private_key_hex, 16)
        tx = await wallet_utils.create_stake_transaction(private_key, amount)
        error_message, transaction_hash = await push_tx(tx, wallet_utils)
        if transaction_hash:
            return transaction_hash
        else:
            raise Exception(error_message)
    except Exception as e:

        raise


async def unstake_transaction(wallet_utils: Utils, private_key_hex):
    try:
        private_key = int(private_key_hex, 16)
        tx = await wallet_utils.create_unstake_transaction(private_key)
        error_message, transaction_hash = await push_tx(tx, wallet_utils)
        if transaction_hash:
            return transaction_hash
        else:
            raise Exception(error_message)
    except Exception as e:

        raise


async def register_inode_transaction(wallet_utils: Utils, private_key_hex):
    try:
        private_key = int(private_key_hex, 16)
        tx = await wallet_utils.create_inode_registration_transaction(private_key)
        error_message, transaction_hash = await push_tx(tx, wallet_utils)
        if transaction_hash:
            return transaction_hash
        else:
            raise Exception(error_message)
    except Exception as e:

        raise


async def deregister_inode_transaction(wallet_utils: Utils, private_key_hex):
    try:
        private_key = int(private_key_hex, 16)
        tx = await wallet_utils.create_inode_de_registration_transaction(private_key)
        error_message, transaction_hash = await push_tx(tx, wallet_utils)
        if transaction_hash:
            return transaction_hash
        else:
            raise Exception(error_message)
    except Exception as e:

        raise


async def register_validator_transaction(wallet_utils: Utils, private_key_hex):
    try:
        private_key = int(private_key_hex, 16)
        tx = await wallet_utils.create_validator_registration_transaction(private_key)
        error_message, transaction_hash = await push_tx(tx, wallet_utils)
        if transaction_hash:
            return transaction_hash
        else:
            raise Exception(error_message)
    except Exception as e:

        raise


async def vote_transaction(
    wallet_utils: Utils, private_key_hex, voting_range, recipient
):
    try:
        private_key = int(private_key_hex, 16)
        tx = await wallet_utils.create_voting_transaction(
            private_key, voting_range, recipient
        )
        error_message, transaction_hash = await push_tx(tx, wallet_utils)
        if transaction_hash:
            return transaction_hash
        else:
            raise Exception(error_message)
    except Exception as e:

        raise


async def revoke_transaction(wallet_utils: Utils, private_key_hex, revoke_from):
    try:
        private_key = int(private_key_hex, 16)
        tx = await wallet_utils.create_revoke_transaction(private_key, revoke_from)
        error_message, transaction_hash = await push_tx(tx, wallet_utils)
        if transaction_hash:
            return transaction_hash
        else:
            raise Exception(error_message)
    except Exception as e:

        raise


async def get_balance(wallet_utils, address: str):
    try:
        balance_info = wallet_utils.get_balance_info(address)
        formatted_total_balance = f"Total Balance: {balance_info[0]}"
        formatted_stake_balance = f"Staked Balance: {balance_info[2]}"

        balance_details = [formatted_total_balance, formatted_stake_balance]

        if balance_info[1] != decimal.Decimal("0"):
            formatted_pending_balance = f"Pending Incoming Balance: {balance_info[1]}"
            balance_details.append(formatted_pending_balance)

        if balance_info[3] != decimal.Decimal("0"):
            formatted_pending_stake_balance = (
                f"Pending Staked Balance: {balance_info[3]}"
            )
            balance_details.append(formatted_pending_stake_balance)

        if balance_info[4]:
            balance_details.append("Balance is syncing...")

        return "\n".join(balance_details)

    except Exception as e:
        raise


async def find_tx(wallet_utils: Utils, hash: str):
    try:
        balance_info = wallet_utils.get_tx(hash)
        return balance_info
    except Exception as e:
        raise


def publickey_to_address(pointx):
    try:
        address = point_to_string(pointx)
        return address
    except Exception as e:
        raise


def address_to_publickey(pointx):
    try:
        address = string_to_point(pointx)
        return address
    except Exception as e:
        raise


def create_wallet():
    """
    Creates a new wallet by generating a private key and deriving its public key.
    Returns a dictionary containing the private key (hex), public key (address), and raw private key.
    """
    try:
        private_key = keys.gen_private_key(CURVE)
        public_key = keys.get_public_key(private_key, CURVE)
        address = point_to_string(public_key)

        return {
            "private_key": hex(private_key),
            "address": address,
            "raw_private_key": private_key,
        }
    except Exception as e:
        raise Exception(f"Failed to create wallet: {str(e)}")



from setuptools import setup, find_packages

setup(
    name="upowpy",
    version="1.0.10",
    author="upow",
    author_email="contact@upow.ai",
    description="upowpy is a Python library for interacting with uPow blockchain transactions, including sending funds, staking, voting, and managing nodes and validators on the blockchain",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/upowai/upowpy",
    packages=find_packages(),
    py_modules=["upowpy"],
    install_requires=[
        "asttokens==2.4.1",
        "base58==0.2.5",
        "certifi==2023.11.17",
        "charset-normalizer==2.0.12",
        "colorama==0.4.6",
        "executing==2.0.1",
        "fastecdsa==2.3.0",
        "icecream==2.1.3",
        "idna==3.6",
        "pygments==2.17.2",
        "requests==2.26.0",
        "six==1.16.0",
        "urllib3==1.26.18",
        "pickleDB~=0.9.2",
    ],
    python_requires=">=3.11",
)



from fastecdsa import curve

ENDIAN = "little"
CURVE = curve.P256
SMALLEST = 100000000
MAX_SUPPLY = 18_884_643
VERSION = 1
MAX_BLOCK_SIZE_HEX = 4096 * 1024  # 4MB in HEX format, 2MB in raw bytes
MAX_INODES = 12



import hashlib
import json
import logging
import sys
from decimal import Decimal
from enum import Enum, IntEnum
from math import ceil
from datetime import datetime, timezone
from typing import Union
import base58
from fastecdsa.point import Point
from fastecdsa.util import mod_sqrt
from icecream import ic

from .constants import ENDIAN, CURVE, SMALLEST

_print = print

logging.basicConfig(
    level=logging.INFO if "--nologs" not in sys.argv else logging.WARNING
)


def log(s):
    logging.getLogger("upow").info(s)


ic.configureOutput(outputFunction=log)


def get_json(obj):
    return json.loads(
        json.dumps(
            obj, default=lambda o: getattr(o, "as_dict", getattr(o, "__dict__", str(o)))
        )
    )


def timestamp():
    return int(datetime.now(timezone.utc).replace(tzinfo=timezone.utc).timestamp())


def sha256(message: Union[str, bytes]):
    if isinstance(message, str):
        message = bytes.fromhex(message)
    return hashlib.sha256(message).hexdigest()


def byte_length(i: int):
    return ceil(i.bit_length() / 8.0)


def normalize_block(block) -> dict:
    block = dict(block)
    block["address"] = block["address"].strip(" ")
    block["timestamp"] = int(
        block["timestamp"].replace(tzinfo=timezone.utc).timestamp()
    )
    return block


def x_to_y(x: int, is_odd: bool = False):
    a, b, p = CURVE.a, CURVE.b, CURVE.p
    y2 = x**3 + a * x + b
    y_res, y_mod = mod_sqrt(y2, p)
    return y_res if y_res % 2 == is_odd else y_mod


class AddressFormat(Enum):
    FULL_HEX = "hex"
    COMPRESSED = "compressed"


class TransactionType(IntEnum):
    REGULAR = 0
    STAKE = 1
    UN_STAKE = 2
    INODE_REGISTRATION = 3
    INODE_DE_REGISTRATION = 4
    VALIDATOR_REGISTRATION = 5
    VOTE_AS_VALIDATOR = 6
    VOTE_AS_DELEGATE = 7
    REVOKE_AS_VALIDATOR = 8
    REVOKE_AS_DELEGATE = 9


class OutputType(IntEnum):
    REGULAR = 0
    STAKE = 1
    UN_STAKE = 2
    INODE_REGISTRATION = 3
    VALIDATOR_REGISTRATION = 5
    VOTE_AS_VALIDATOR = 6
    VOTE_AS_DELEGATE = 7
    VALIDATOR_VOTING_POWER = 8
    DELEGATE_VOTING_POWER = 9


class InputType(IntEnum):
    REGULAR = 0
    FEES = 1


def get_transaction_type_from_message(message: bytes) -> TransactionType:
    # Create a reverse mapping from string to enum value
    transaction_type_mapping = {str(item.value): item for item in TransactionType}

    try:
        # Convert bytes to string
        str_message = simple_bytes_to_string(message)

        # Convert str_message back to enum
        decoded_message = int(str_message)
        transaction_type = transaction_type_mapping.get(
            str(decoded_message), TransactionType.REGULAR
        )
    except (UnicodeDecodeError, ValueError, TypeError):
        # Handle potential decoding errors or value errors
        transaction_type = TransactionType.REGULAR

    return transaction_type


def simple_bytes_to_string(data: bytes) -> str:
    if data is None:
        return None
    try:
        # Attempt to decode bytes as hex
        return data.decode("utf-8")
    except UnicodeDecodeError:
        # If decoding as hex fails, assume it's already a string
        return data.hex()


def point_to_bytes(
    point: Point, address_format: AddressFormat = AddressFormat.FULL_HEX
) -> bytes:
    if address_format is AddressFormat.FULL_HEX:
        return point.x.to_bytes(32, byteorder=ENDIAN) + point.y.to_bytes(
            32, byteorder=ENDIAN
        )
    elif address_format is AddressFormat.COMPRESSED:
        return string_to_bytes(point_to_string(point, AddressFormat.COMPRESSED))
    else:
        raise NotImplementedError()


def bytes_to_point(point_bytes: bytes) -> Point:
    if len(point_bytes) == 64:
        x, y = int.from_bytes(point_bytes[:32], ENDIAN), int.from_bytes(
            point_bytes[32:], ENDIAN
        )
        return Point(x, y, CURVE)
    elif len(point_bytes) == 33:
        specifier = point_bytes[0]
        x = int.from_bytes(point_bytes[1:], ENDIAN)
        return Point(x, x_to_y(x, specifier == 43))
    else:
        raise NotImplementedError()


def round_up_decimal(decimal: Decimal, round_up_length: str = "0.00000001"):
    round_up_length = Decimal(round_up_length)
    if (decimal * SMALLEST) % 1 != 0.0:
        decimal = decimal.quantize(round_up_length)
    return decimal


def bytes_to_string(point_bytes: bytes) -> str:
    point = bytes_to_point(point_bytes)
    if len(point_bytes) == 64:
        address_format = AddressFormat.FULL_HEX
    elif len(point_bytes) == 33:
        address_format = AddressFormat.COMPRESSED
    else:
        raise NotImplementedError()
    return point_to_string(point, address_format)


def point_to_string(
    point: Point, address_format: AddressFormat = AddressFormat.COMPRESSED
) -> str:
    if address_format is AddressFormat.FULL_HEX:
        point_bytes = point_to_bytes(point)
        return point_bytes.hex()
    elif address_format is AddressFormat.COMPRESSED:
        x, y = point.x, point.y
        address = base58.b58encode(
            (42 if y % 2 == 0 else 43).to_bytes(1, ENDIAN) + x.to_bytes(32, ENDIAN)
        )
        return address if isinstance(address, str) else address.decode("utf-8")
    else:
        raise NotImplementedError()


def string_to_bytes(string: str) -> bytes:
    try:
        point_bytes = bytes.fromhex(string)
    except ValueError:
        point_bytes = base58.b58decode(string)
    return point_bytes


def string_to_point(string: str) -> Point:
    return bytes_to_point(string_to_bytes(string))
